# class K for real-time K

import sys, time
import math
import numpy as np
from collections import deque

# prevent Python from writing .pyc files
sys.dont_write_bytecode = True

from point import Point
from fixation import Fixation
from bw import Butterworth as BW 
from sg import SG
from monitor import Monitor
from statistics import mean, stdev, median

class K:
  
    def __init__(self):

        self.d_bar = 0.0            # current fixation duration
        self.a_bar = 0.0            # current saccade amplitude

        self.d_mu = 0.0             # windowed fixation duration mean
        self.d_sigma = 0.0          # windowed fixation duration sdev
        self.d_k = 0                # windowed fixation duration counter

        self.a_mu = 0.0             # windowed amplitude mean
        self.a_sigma = 0.0          # windowed amplitude sdev
        self.a_k = 0                # windowed amplitude counter

        # this "memory" will have a large effect on how quickly K gets
        # updated (cleared out)
        # let's assume rhythmic theory of attention: 2-8 Hz, meaning
        # 125-500 ms sampling period, maybe set to 125?
        self.win = 125
#       self.win = 66
#       self.win = 33
#       self.win = 30

        self.K = 0.0		        # current K coefficient
        self.Kbw = 0.0		        # current K coefficient (smoothed)
        self.Kring = deque(np.zeros(self.win),maxlen=self.win)

        self.pts = deque(maxlen=self.win) # window of past points

        self.dur = deque(maxlen=self.win) # window of fixation durations
        self.amp = deque(maxlen=self.win) # window of saccade amplitudes

        self.bwf_x = BW(2,30,1.15)  # Butterworth filter init
        self.bwf_y = BW(2,30,1.15)  # Butterworth filter init

        self.sgf_x = SG(5,2,1)      # Savitzky-Golay init (width,degree,op)
        self.sgf_y = SG(5,2,1)      # Savitzky-Golay init (width,degree,op)

        self.state = 0              # 0: saccade, 1: fixation
        self.threshold = 30.0       # velocity threshold (30 deg/s)

        # display arguments: x,y,w,h,dist,n, window size we want
        self.monitor = Monitor(0,0,1680,1050,22,19.68)

    def window(self):
        return self.win

    def update(self):

        # update K, duration - amplitude, normalized z-scores
        if self.d_sigma > 0.0001 and self.a_sigma > 0.0001:
            self.K = (self.d_bar - self.d_mu)/self.d_sigma - \
                     (self.a_bar - self.a_mu)/self.a_sigma
            self.Kbw = self.bwf_x.bwlpf(self.K)
#           self.Kring.append(self.Kbw)
            self.Kring.append(self.squash(self.Kbw))
#           self.Kring.append(self.squash(self.K))
            print("smootehd K %f" % (self.Kbw))

        if len(self.dur) > 1 and len(self.amp):
            mdur = mean(self.dur)
            mamp = mean(self.amp)
            print("mean duration %f, mean amplitude %f" % (mdur, mamp))

            ddur = median(self.dur)
            damp = median(self.amp)
            print("median duration %f, median amplitude %f" % (ddur, damp))

            sdur = math.sqrt(stdev(self.dur))
            samp = math.sqrt(stdev(self.amp))
            sdur = 0.01 if abs(sdur) < 0.0001 else sdur
            samp = 0.01 if abs(samp) < 0.0001 else samp
            print("stdev duration %f, stdev amplitude %f" % (sdur, samp))

            altk = (self.d_bar - mdur)/sdur - \
                   (self.a_bar - mamp)/samp
            print(
                "windowed K %f, squashed windowed K %f" %
                (altk, self.squash(altk))
            )

#       return self.K
#       return self.squash(self.K)
        return self.Kring[-1]

    def squash(self, v):
        # use logistic function to bound (squash) K response

        # the supremum of the values of the function range between
        # asymptotes [-1:1]
        L = 2.0

        # logistic growth rate or steepness of the curve (bounds x values, e.g.,
        # if k=2, the clamping allos K to range [-2:2]
#       k = 2.0
        # this parameter depends on what the expected range of K is...if it
        # goes [-100:100] then maybe 0.1 makes sense...but I've no idea what
        # the expected rnage of K should be...
        # let's assume rhythmic theory of attention: 2-8 Hz, meaning
        # 125-500 ms sampling period, maybe set to .125?
        k = 0.125
#       k = 0.250
        try:
#           logis = L/(1. + math.exp(-k * (v))) - 1
            logis = L/(1. + np.exp(-k * (v))) - 1
        except OverflowError:
            print("OverflowError: k: %f, v: %f" % (k, v))
            logis = 0.0
        return logis

    def smooth(self,x,y,t):
        # smooth points
        bwx = self.bwf_x.bwlpf(x)
        bwy = self.bwf_y.bwlpf(y)

        return (bwx, bwy)

    def filter(self,x,y,t,dt):
        # t: current timestamp, dt: time period between samples (ms)

        # velocity in screen coords (normalized)
        dx, dy = self.sgf_x.sgf(x), self.sgf_y.sgf(y)

        # sampling period in sec
        period = float(dt)/1000.0
        
        # multiply by sgf length
        dt = period * float(self.sgf_x.n)

        # use monitor to compute degrees, note monitor uses screen dimensions
        deg_x, deg_y = self.monitor.pix2deg(dx), self.monitor.pix2deg(dy)

        # deg_x, deg_y are in deg per filter window, div by dt for per sec
        vel_x = deg_x / dt
        vel_y = deg_y / dt

        vel = math.sqrt(vel_x*vel_x + vel_y*vel_y)

        # print("velocity (pixels) = (%f,%f)" % (dx,dy))
        # print("velocity (deg) = (%f,%f)" % (velx,vely))
        # print("velocity (deg) = (%f)" % (vel))

        return vel

    def classify(self,x,y,t,dt,vel):

        # state table
        # state         |   input    |  output
        # ------------------------------------
        # (1) fixation  | a < T (1)  |    (1) fixation
        # (0) saccade   | a < T (1)  |    (1) fixation
        # (1) fixation  | a > T (0)  |    (0) saccade
        # (0) saccade   | a > T (0)  |    (0) saccade
        #

        # amplitude > T (saccade)
        if math.fabs(vel) > self.threshold:
            if self.state == 1:
                # state transition frmo fixation to saccade
                self.fxtn_sacc(x,y,t,dt)
            else:
                # state transition frmo saccade to saccade
                self.sacc_sacc(x,y,t,dt)
            self.state = 0
        # amplitude < T (fixation)
        else:
            if self.state == 1:
                # state transition frmo fixation to fixation
                self.fxtn_fxtn(x,y,t,dt)
            else:
                # state transition frmo saccade to fixation
                self.sacc_fxtn(x,y,t,dt)
            self.state = 1

        # add point to short-term memory (deque)
        self.pts.append(Point(x,y,t,None))

        return self.state

    def sacc_sacc(self,x,y,t,dt):
        # transition from saccade to saccade

        # accumulate dist of current gaze point to previous gaze point
        if len(self.pts):
            # compute distance to last point
            dx, dy = x - self.pts[-1].x, y - self.pts[-1].y
            # convert to degrees visual angle
            deg_x, deg_y = self.monitor.pix2deg(dx), self.monitor.pix2deg(dy)
            # compute distance (deg visual angle)
            dis = math.sqrt(deg_x*deg_x + deg_y*deg_y)
            self.a_bar += dis

            # store amplitude in deque
            self.amp.append(dis)

            # update mean, sdev of saccade amplitude 
            self.a_mu = float(self.a_k)/float(self.a_k + 1) * self.a_mu + \
                1.0/float(self.a_k + 1) * dis

            if self.a_k > 0:
                self.a_sigma = math.sqrt(
                    float(self.a_k-1)/float(self.a_k) * self.a_sigma**2 + \
                                  1.0/float(self.a_k+1) * (dis-self.a_mu)**2
                    )
            self.a_k += 1

    def fxtn_fxtn(self,x,y,t,dt):
        # transition from fixation to fixation

        # accumulate time delta of current gaze point to previous gaze point
        if len(self.pts):
            # compute time since last point
            dt = t - self.pts[-1].t
            # convert to seconds
            dur = dt/1000.0
            self.d_bar += dur

            # update mean, sdev of fixation duration 
            self.d_mu = float(self.d_k)/float(self.d_k + 1) * self.d_mu + \
                1.0/float(self.d_k + 1) * dur

            if self.d_k > 0:
                self.d_sigma = math.sqrt(
                    float(self.d_k-1)/float(self.d_k) * self.d_sigma**2 + \
                                  1.0/float(self.d_k+1) * (dur-self.d_mu)**2
                    )
            self.d_k += 1

    def sacc_fxtn(self,x,y,t,dt):
        # transition from saccade to fixation

        # saccade ends, store this saccade's amplitude in deque
        self.amp.append(self.a_bar)

        # current saccade ends, new fixation begins, reset fixation duration
        self.d_bar = 0.0
        # do we need this?
        self.d_mu = 0.0             # windowed fixation duration mean
        self.d_sigma = 0.0          # windowed fixation duration sdev
        self.d_k = 0                # windowed fixation duration counter

    def fxtn_sacc(self,x,y,t,dt):
        # transition from fixation to saccade

        # fixation ends, store this fixation's duration in deque
        self.dur.append(self.d_bar)

        # current fixation ends, new saccade begins, reset saccade amplitude
        self.a_bar = 0.0
        # do we need this?
        self.a_mu = 0.0             # windowed amplitude mean
        self.a_sigma = 0.0          # windowed amplitude sdev
        self.a_k = 0                # windowed amplitude counter
