#!/usr/bin/env python3

import platform, sys, os, math
import argparse
import numpy as np
import time
import datetime
from collections import deque
from PIL import Image
import cv2

from state import State
from point import Point
from fixation import Fixation
from bw import Butterworth as BW
from sg import SG
from monitor import Monitor
from calibration import Calibration
from gazept import Gazept
from rstats import RunningStats
from kcoeff import K

from math import *
from OpenGL.GL import *
from OpenGL.GLUT import *
from OpenGL.GLU import *

print(OpenGL.__file__)

# prevent Python from writing .pyc files
sys.dont_write_bytecode = True

# globals
global monitor, state, calibration, gazept
global bw_filter_x, bw_filter_y
global sg_filter_x, sg_filter_y
global k_coeff
global generate_video, video_out

def clamp(x,y):
    if x < 0:
      x = 0.0
    if y < 0:
      y = 0.0
    if x > 1:
      x = 1.0
    if y > 1:
      y = 1.0
    return x,y

def idle():
  global monitor, state
  global gazept

  if gazept is not None:
    if state.get_input() == 1 and gazept.get_state() == "receiving":
      gaze_motion()

# glutPostRedisplay()

def timer(extra):
  glutPostRedisplay()

  # re-arm timer
  glutTimerFunc(60,timer,0)

def reshape(width, height):
  # only call glViewport once here, not in display()!
  glViewport(0, 0, width, height)

  glMatrixMode(GL_PROJECTION)
  glLoadIdentity()

  glMatrixMode(GL_MODELVIEW)
  glLoadIdentity()

# monitor.set_window_width(glutGet(GLUT_WINDOW_WIDTH))
# monitor.set_window_height(glutGet(GLUT_WINDOW_HEIGHT))

def display():
  global monitor, state
  global gazept
  global k_coeff
  global generate_video, video_out

  # render

  ww = glutGet(GLUT_WINDOW_WIDTH)
  wh = glutGet(GLUT_WINDOW_HEIGHT)

  glClearColor(1.0, 1.0, 1.0, 0.5)
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

  # draw K
  # glViewport(x,y,width,height)
# glViewport(0, 7*wh//100, ww, 11*wh//100)
  glViewport(0, 5*wh//100, ww, 13*wh//100)
  glMatrixMode(GL_PROJECTION)
  glLoadIdentity()

  # l,r,b,t,znear,zfar
  glOrtho(0, k_coeff.window(),  -1.1, 1.1, -1.0, 1.0)

  glMatrixMode(GL_MODELVIEW)
  glLoadIdentity()

  # grey (zero line)
  glColor4f(192.0/255.0, 192.0/255.0, 192.0/255.0, 1.0)
  glBegin(GL_LINE_STRIP)
  for x, v in enumerate(k_coeff.Kring):
    glVertex3f(x,0.0,0.0)
  glEnd()

  # yellow
# glColor4f(255.0/255.0, 247.0/255.0, 188.0/255.0, 1.0)
  # oranges
  glColor4f(230.0/256.0,  85.0/256.0,  13.0/256.0, 0.6)
  glBegin(GL_LINE_STRIP)
  for x, v in enumerate(k_coeff.Kring):
    glVertex3f(x,float(v),0.0)
  glEnd()

  # yellow
# glColor4f(255.0/255.0, 247.0/255.0, 188.0/255.0, 0.6)
  # oranges
  glColor4f(230.0/256.0,  85.0/256.0,  13.0/256.0, 0.6)
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
  glBegin(GL_TRIANGLE_STRIP)
  for x, v in enumerate(k_coeff.Kring):
    glVertex3f(x,float(v),0.0)
    glVertex3f(x,0.0,0.0)
  glEnd()
  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)

  glViewport(0, 0, ww, wh)
  glMatrixMode(GL_PROJECTION)
  glLoadIdentity()

  glOrtho(0.0, ww, 0.0, wh, 0.0, 1.0)

  glMatrixMode(GL_MODELVIEW)
  glLoadIdentity()

  if gazept is not None and gazept.get_state() == "calibrating":

    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)

    v = calibration.get_xy()
    cx = v[0] * ww
    cy = v[1] * wh
    # do the y-flip
    cy = wh - cy
    r = calibration.get_dot_radius()
#   print "dot position (pixels) = (%d,%d)" % (cx,cy)
#   print "radius = %f" % (r)
    glColor3f(1.0, 1.0, 1.0)
    draw_circ(cx, cy, r, 36, GL_TRIANGLE_FAN)

    # advance calibration timer
    do_calibration()

  else:

    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)

    if state.get_input() == 1:
      if gazept is not None and gazept.get_state() == "receiving":
        x = state.get_gaze_coord_x() * ww
        y = state.get_gaze_coord_y() * wh
      else:
        x = 0.5 * ww
        y = 0.5 * wh

      # render K beside the gaze point
      if state.K < 0:
        glColor3f(49.0/256.0, 130.0/256.0, 247.0/256.0)
      else:
        glColor3f(230.0/256.0, 85.0/256.0, 13.0/256.0)
      font = GLUT_BITMAP_HELVETICA_18
      glRasterPos3f(x+10,wh-(y+10),0.0)
      infostr = "%2.1f" % (state.K)
      for ch in infostr:
        glutBitmapCharacter(font, ctypes.c_int(ord(ch)))

    else:
      x = state.get_mouse_coord_x() * ww
      y = state.get_mouse_coord_y() * wh

      # render mouse coords beside point
      glColor3f(1.0, 1.0, 1.0)
      font = GLUT_BITMAP_HELVETICA_18
      glRasterPos3f(x+10,wh-(y+10),0.0)
      infostr = "%d %d" % (x,y)
      for ch in infostr:
        glutBitmapCharacter(font, ctypes.c_int(ord(ch)))

    # pass (x,y) through filter/threshold function
    x, y = filter(x,y,16)

    t = glutGet(GLUT_ELAPSED_TIME)

    # compute K
    vel = k_coeff.filter(x,y,t,16)
    state.set_fixating(k_coeff.classify(x,y,t,16,vel))
    state.K = k_coeff.update()

    # set color and then draw fixation rectangle if fixating
#   rect_dim = 20 if state.get_fixating() else 10
    if state.get_fixating():
      # colorbrewer blues
      glColor3f(49.0/256.0, 130.0/256.0, 247.0/256.0)

      # draw fixation as a rectangle in screen coordinates
#     rect_dim = 10
#     draw_rect(state.ux,wh-state.uy,rect_dim,rect_dim)
      draw_rect(state.ux,wh-state.uy,state.sx,state.sy)

      # this was testing running stats object
#     glColor3f(1.0, 1.0, 0.0)
#     draw_rect(state.rx.avg(),wh-state.ry.avg(),state.rx.sdv(),state.ry.sdv())
    else:
      # colorbrewer oranges
      glColor3f(230.0/256.0, 85.0/256.0, 13.0/256.0)


    # render gaze point (buffered points are normalized)
#   glColor3f(1.0, 1.0, 1.0)
#   draw_circ(x, wh - y, 5, 18, GL_TRIANGLE_FAN)
    glDrawCircle(x, wh - y, 5, 18)

    dk = len(state.points) - 1
    while dk > 1:

      # y-flip
#     glDrawCircle(state.points[dk].x*ww, (1. - state.points[dk].y)*wh, 3, 36)
      glDrawCircle(state.points[dk].x*ww, (1. - state.points[dk].y)*wh, 5, 18)

      glBegin(GL_LINES)
      # y-flip
      glVertex2f(state.points[dk].x*ww, (1. - state.points[dk].y)*wh)
      glVertex2f(state.points[dk-1].x*ww, (1. - state.points[dk-1].y)*wh)
      glEnd()

      dk -= 1

  # get current frame time (ms)
  state.vms = glutGet(GLUT_ELAPSED_TIME)
  state.vtd = datetime.timedelta(milliseconds=state.vms)

  # text: HH:MM:SS,mmmmmm
# glColor4f(1.0, 1.0, 1.0, 0.9)
  glColor4f(0.0, 0.0, 0.0, 0.9)
# font = GLUT_BITMAP_9_BY_15
  font = GLUT_BITMAP_HELVETICA_18
  hmsstr = str(state.vtd).replace('.',',')
# hmsstr = str(state.vtd)
  infostr = "%s" % (hmsstr)
  glRasterPos3f(10.0,10.0,0.0)
  for ch in infostr:
    glutBitmapCharacter(font, ctypes.c_int(ord(ch)))


  # text: duration of current fixation
  glColor4f(1.0, 1.0, 1.0, 0.9)
  if state.get_fixating():
    # use len(infostr) of infostr from above)
    glRasterPos3f(len(infostr)*10.0,10.0,0.0)
    infostr = "%s" % (state.tt)
    for ch in infostr:
      glutBitmapCharacter(font, ctypes.c_int(ord(ch)))

  # text: K
  if state.K < 0:
    glColor3f(230.0/256.0, 85.0/256.0, 13.0/256.0)
  else:
    glColor3f(49.0/256.0, 130.0/256.0, 247.0/256.0)
  infostr = "K = %2.1f" % (state.K)
  glRasterPos3f(ww - len(infostr)*10.0,10.0,0.0)
  for ch in infostr:
    glutBitmapCharacter(font, ctypes.c_int(ord(ch)))

# font = GLUT_STROKE_ROMAN
# glRasterPos3f(10.0,10.0,0.0)
# for ch in "Hello World":
#   glutStrokeCharacter(font, ctypes.c_int(ord(ch)))

  glutSwapBuffers()

  if generate_video:
    glReadBuffer(GL_FRONT)
    buf = glReadPixels(0,0,ww,wh,GL_RGBA,GL_UNSIGNED_BYTE)
    img = Image.frombytes(mode="RGBA",size=(ww,wh),data=buf)
    frame = cv2.cvtColor(np.asarray(img),cv2.COLOR_RGB2BGR)
    frame = cv2.flip(frame,0)
    video_out.write(frame)

def do_calibration():
  global gazept
  global calibration

  # moving dot
  if calibration.movement == 0:
    t = calibration.timer_elapsed_ms()/calibration.get_moveTime()
    i = calibration.get_dot_index()
    if i+1 < calibration.get_dot_len():
      v1 = calibration.get_dot(i)
      v2 = calibration.get_dot(i+1)
      t = 60.0*((1.0/10.0)*t**5 - (1.0/4.0)*t**4 + (1.0/6.0)*t**3)
      v = ((1.0 - t)*v1[0] + t*v2[0], (1.0 - t)*v1[1] + t*v2[1])
      calibration.set_xy(v)
    if calibration.timer_elapsed_ms() > calibration.get_moveTime():
      calibration.inc_dot_index()
      calibration.timer_start()
      # advance to dilation
      calibration.movement += 1

  # dilating dot
  elif calibration.movement == 1:
    t = calibration.timer_elapsed_ms()/calibration.get_zoomTime()
    r = calibration.get_innerR() + t * \
       (calibration.get_outerR() - calibration.get_innerR())
    calibration.set_dot_radius(r)
    v = (calibration.get_dot_coord_x(calibration.get_dot_index()), \
         calibration.get_dot_coord_y(calibration.get_dot_index()))
    calibration.set_xy(v)
    if calibration.timer_elapsed_ms() > calibration.get_zoomTime():
      calibration.timer_start()
      # advance to shrinking
      calibration.movement += 1

  # shrinking dot
  elif calibration.movement == 2:
    t = calibration.timer_elapsed_ms()/calibration.get_zoomTime()
    r = calibration.get_outerR() - t * \
       (calibration.get_outerR() - calibration.get_innerR())
    calibration.set_dot_radius(r)
    v = (calibration.get_dot_coord_x(calibration.get_dot_index()), \
         calibration.get_dot_coord_y(calibration.get_dot_index()))
    calibration.set_xy(v)
    if calibration.timer_elapsed_ms() > calibration.get_zoomTime():
      i = calibration.get_dot_index()
      if i+1 == calibration.get_dot_len():
        calibration.inc_dot_index()
        # calibration ends here
        if gazept is not None:
          gazept.stop_calibration()
        calibration.timer_end()
      else:
        calibration.timer_start()
      # reset to movement
      calibration.movement = 0

  else:
    pass

  # override 60 fps timer to get smoother calibration animation
# glutPostRedisplay()

def draw_rect(x, y, width, height):
  # draw in screen coordinates
  glBegin(GL_QUADS)
  glVertex2f(x - (width / 2), y - (height / 2))
  glVertex2f(x + (width / 2), y - (height / 2))
  glVertex2f(x + (width / 2), y + (height / 2))
  glVertex2f(x - (width / 2), y + (height / 2))
  glEnd()

def draw_circ(cx, cy, r, num_segments, type):
  theta = 2 * 3.1415926 / float(num_segments)
  # tangential factor
  tanf = tan(theta)
  # radial factor
  radf = cos(theta)
  # start x at angle 0
  x = r
  y = 0
  glBegin(type)
  i = 0

  for i in range(num_segments):
    glVertex2f(x + cx, y + cy)
    tx = -y
    ty = x
    x = x + tx * tanf
    y = y + ty * tanf
    x = x * radf
    y = y * radf

  glEnd()

def glDrawCircle(cx,cy,radius,slices):

  glPushMatrix()
  glTranslatef(cx,cy,0.0)

# glBegin(GL_POLYGON)
# for v in xrange(0, slices):
#   angle  = float(v) * 2.0 * np.pi / slices
#   glVertex2f(np.cos(angle)*radius, np.sin(angle)*radius)
# glEnd()

  thickness = 1

  quadric = gluNewQuadric()
  gluDisk(quadric,radius-thickness,radius,slices,1)
  del quadric

  glPopMatrix()

def glDrawDisc(cx,cy,radius,slices):

  glPushMatrix()
  glTranslatef(cx,cy,0.0)

# glBegin(GL_POLYGON)
# for v in xrange(0, slices):
#   angle  = float(v) * 2.0 * np.pi / slices
#   glVertex2f(np.cos(angle)*radius, np.sin(angle)*radius)
# glEnd()

  quadric = gluNewQuadric()
  gluDisk(quadric,0,radius,slices,1)
  del quadric

  glPopMatrix()

def keyboard(bkey, x, y):
  global monitor, state
  global gazept
  global generate_video, video_out

  key = bkey.decode("utf-8")

  if key == '\033' or key == 'q':
    # if eye tracker is in use, shut it down (else we may hang)
    if gazept is not None and gazept.get_state() != "shutdown":
      gazept.shutdown()
    if generate_video:
      video_out.release()
    os._exit(0)
  # enter and exit butterworth filter mode
  if key == 'b':
    print("toggling butterworth: ", "on" if state.get_butterworth() else "off")
    state.toggle_butterworth()
  # run or record eye tracking data
  if key == 'r':
    if state.get_input():
      if gazept is not None and gazept.get_state() != "calibrating":
        gazept.start_receiving()
  # set interface to calibration
  if key == 'c':
    if state.get_input():
      if gazept is not None and gazept.get_state() != "calibrating":

        # clear out calibration points, need to resend coords and timing
        gazept.calibrate_clear()
        for i, point in enumerate(calibration.get_calibration_points()):
          # skip the first (origin) point
          if i > 0:
            gazept.calibrate_addpoint(point[0], point[1])
        # need twice the zoomTime since we dilate/constrict the calib point
        gazept.calibrate_timeout(2.0*calibration.get_zoomTime()/1000.0)
        gazept.calibrate_delay(calibration.get_moveTime()/1000.0)

        gazept.start_calibration()
      calibration.timer_start()
  # toggle input
  if key == 'i':
    if state.get_input() == -1:
      print("input: ", "mouse")
    else:
      state.set_input(1 - state.get_input())
      print("input: ", "eye tracker" if state.get_input() else "mouse")
  # toggle fullscreen
  if key == 'f':
    state.toggle_fullscreen()
    wx = monitor.get_window_pos_x()
    wy = monitor.get_window_pos_y()
    sw = monitor.get_fullscreen_width()
    sh = monitor.get_fullscreen_height()
    ww = monitor.get_window_width()
    wh = monitor.get_window_height()
    if state.get_fullscreen():
      glutFullScreen()
      if gazept is not None:
        gazept.screen_size(wx,wy,sw,sh)
    else:
      glutReshapeWindow(ww,wh)
      glutPositionWindow(wx,wy)
      if gazept is not None:
        gazept.screen_size(wx,wy,ww,wh)

# glutPostRedisplay()

def mouse_motion(mx, my):
  global monitor, state

  ww = glutGet(GLUT_WINDOW_WIDTH)
  wh = glutGet(GLUT_WINDOW_HEIGHT)

  # get normalized coords
  x = float(mx)/ww
  y = float(my)/wh

  # update mouse position (received in screen coords)
  state.set_mouse_coord_x(x)
  state.set_mouse_coord_y(y)

# print("mouse motion (%f x %f)" % (float(x),float(y))

def gaze_motion():
  global monitor, state
  global gazept

  if gazept is not None:
#   x = gazept.get_gaze_coord_x()
#   y = gazept.get_gaze_coord_y()
    x,y = gazept.get_gaze_coord()

    # update gaze position (received in normalized coords)
    state.set_gaze_coord_x(float(x))
    state.set_gaze_coord_y(float(y))

# print("gaze motion (%f x %f)" % (float(x),float(y)))

def filter(x,y,dt):
  global monitor, state
  global bw_filter_x, bw_filter_y
  global sg_filter_x, sg_filter_y
  global k_coeff

  ww = glutGet(GLUT_WINDOW_WIDTH)
  wh = glutGet(GLUT_WINDOW_HEIGHT)

  # smooth points
  if state.get_butterworth() == 1:
    bwx = bw_filter_x.bwlpf(x)
    bwy = bw_filter_y.bwlpf(y)
    if state.get_input() == 1:
      state.set_gaze_coord_x(bwx / ww)
      state.set_gaze_coord_y(bwy / wh)
    else:
      state.set_mouse_coord_x(bwx / ww)
      state.set_mouse_coord_y(bwy / wh)
    # if filtering is on, reset x,y for velocity filtering below
    x = bwx
    y = bwy

  # add point to deque
  # need to put this here since we might want to enqueue smoothed points
  t = glutGet(GLUT_ELAPSED_TIME)
  state.points.append(Point(x/ww,y/wh,t,None))

  return (x,y)

# glutPostRedisplay()

def main(argv):
  # init global state, calibration
  global monitor, state, calibration, gazept
  global bw_filter_x, bw_filter_y
  global sg_filter_x, sg_filter_y
  global k_coeff
  global generate_video, video_out

  generate_video = False

  parser = argparse.ArgumentParser(description='Real-time K simulator')
  parser.add_argument('--generate_video', action='store_true', help='Generate output video')

  args = parser.parse_args()

  if 'generate_video' in args:
    generate_video = args.generate_video

  # state (global variables)
  state = State()

  # K coeff
  k_coeff = K()

  # eye tracker
  gazept = None

  state.vtd = datetime.timedelta()

  # set initial parameters for eye movement analysis and initial coords
  state.set_mouse_coord_x(.5)
  state.set_mouse_coord_y(.5)
  state.set_dfwidth(5)
  state.set_dfdegree(2)
  state.set_dfo(1)
  state.set_threshold(10)
  state.set_butterworth(1)
# state.set_fullscreen(1)
  state.set_fullscreen(0)

  # set up filters
# bw_filter_x = BW(2, 60, 6.15)
# bw_filter_y = BW(2, 60, 6.15)
  bw_filter_x = BW(2, 60, 9.15)
  bw_filter_y = BW(2, 60, 9.15)
  sg_filter_x = SG(state.get_dfwidth(), state.get_dfdegree(), state.get_dfo())
  sg_filter_y = SG(state.get_dfwidth(), state.get_dfdegree(), state.get_dfo())

  glutInit(argv)
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_ALPHA | GLUT_DEPTH)

  # get screen, window dimensions and position
  sw = glutGet(GLUT_SCREEN_WIDTH)
  sh = glutGet(GLUT_SCREEN_HEIGHT)
  # get desired window dimensions
  ww = 840
  wh = 525
  # figure out desired window position
  wx = (sw - ww)//2
  wy = (sh - wh)//2
  if sys.platform == 'win32':
    # subtract width of window decorations (Windows)
    wx -= 8
    wy -= 30

  print("screen res = ", sw, ' x ', sh)
  print("window res = ", ww, ' x ', wh)
  print("window pos = ", wx, ' x ', wy)

  size = (ww, wh)
  if generate_video:
      print("outputting to 'output.mov'")
      fourcc = cv2.VideoWriter_fourcc(*"mp4v")
      fps = 30.0
      video_out = cv2.VideoWriter("./output.mov", fourcc, fps, size, True)

  # display arguments: x,y,w,h,dist,n, this is window size we want if not full
# monitor = Monitor(0,0,512,512,22,19.68)
# monitor = Monitor(0,0,1680,1050,22,19.68)
  monitor = Monitor(wx,wy,ww,wh,sw,sh,27,19.68)

  # get calibration object
  calibration = Calibration()

  try:
#   gazept = Gazept()
#   gazept = Gazept(server='127.0.0.1', port=4242, debug=False)
#   gazept = Gazept(server='192.168.1.114', port=4242, debug=True)
#   gazept = Gazept(server='198.21.250.230', port=4242, debug=True)
#   gazept = Gazept(server='198.21.227.146', port=4242, debug=True)
#   gazept = Gazept(server='198.21.230.221', port=4242, debug=True)
    gazept = Gazept(server='130.127.48.164', port=4242, debug=True)

    # set up eye-tracking screen size
    if state.get_fullscreen():
      gazept.screen_size(wx,wy,sw,sh)
    else:
      gazept.screen_size(wx,wy,ww,wh)

    # set up calibration coordinates, clear calib points, turn off calib display
    gazept.calibrate_show(0)
    gazept.calibrate_clear()
    for i, point in enumerate(calibration.get_calibration_points()):
      # skip the first (origin) point
      if i > 0:
        gazept.calibrate_addpoint(point[0], point[1])
    # need twice the zoomTime since we dilate/constrict the calib point
    gazept.calibrate_timeout(2.0*calibration.get_zoomTime()/1000.0)
    gazept.calibrate_delay(calibration.get_moveTime()/1000.0)
    # don't start receiving data until after calibration
#   gazept.start_receiving()

    # set up what we want to receive
    gazept.send_counter(1)	# turns on/off counter
    gazept.send_cursor(0)         # turns on/off cursor
    gazept.send_eye_left(0)	# turns on/off eye left (3D data)
    gazept.send_eye_right(0)	# turns on/off eye right (3D data)
    gazept.send_pog_best(1)	# turns on/off POG best
    gazept.send_pog_fix(0)	# turns on/off POG fix
    gazept.send_pog_left(1)	# turns on/off POG left
    gazept.send_pog_right(1)	# turns on/off POG right
    gazept.send_pupil_left(1)	# turns on/off pupil left
    gazept.send_pupil_right(1)	# turns on/off pupil right
    gazept.send_time(1)           # turns on/off time
    gazept.send_time_tick(0)	# turns on/off timetick
    gazept.send_user_data(0)	# turns on/off user data
    gazept.user_data(0)		# set value of user data (usually string)
    gazept.tracker_display(0)	# shows or hides tracker (Control) display

  except:
    state.set_input(-1)

  window = glutCreateWindow('Gazepoint client demo: press Esc to exit')
# if sys.platform == 'win32':
#   glutFullScreen()
# elif sys.platform == "darwin":
#   glutReshapeWindow(ww,wh)
#   glutPositionWindow(wx,wy)
#   glutReshapeFunc(reshape)
# else:
#   print "Unknown system"
#   sys.exit()
  # call of these so that they can be coalesced
  # (glutFullScreen  works on the Mac provided freeglut is not active)
  if state.get_fullscreen():
    glutFullScreen()
  else:
    glutReshapeWindow(ww,wh)
    glutPositionWindow(wx,wy)

  # make window current
  glutSetWindow(window)

  # set up callbacks
  glutReshapeFunc(reshape)
  glutDisplayFunc(display)
  glutIdleFunc(idle)
  glutKeyboardFunc(keyboard)
  glutPassiveMotionFunc(mouse_motion)
  glutTimerFunc(60,timer,0)

  # GL initialization
  glEnable(GL_LINE_SMOOTH)
  glShadeModel(GL_FLAT)

  # blending function
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
  glEnable(GL_BLEND)

# glPolygonMode(GL_FRONT_AND_BACK, GL_FILL)
  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)

  glutMainLoop()

#################################
if __name__ == '__main__':
  main(sys.argv)
